from dataclasses import dataclass

@dataclass
class ProgramNode:
    body: list

    def __repr__(self):
        return (f"PROGRAM({self.body})")

@dataclass
class BinaryOpNode:
    left: any
    op: any
    right: any
    

    def __repr__(self):
        return (f"({self.left} {self.op.value} {self.right})")

@dataclass
class NothingLiteralNode:
    literal: None

    def __repr__(self):
        return (f"Nothing") if self.literal == None else ("(ERROR_VALUE)")

@dataclass
class LiteralNode:
    literal: any

    def __repr__(self):
        return str(self.literal.value)
    
@dataclass
class UnaryOpNode:
    op: any
    operand: any

    def __repr__(self):
        return (f"{self.op.value}{self.operand}")

@dataclass
class PostfixOpNode: ###
    identifier: str
    op: any

    def __repr__(self):
        return (f"({self.identifier}{self.op.value})")
    
@dataclass
class ArrayNode:
    elements: list

    def __repr__(self):
        return f"ARRAY({self.elements})"
    
@dataclass
class IndexAccessNode:
    array: any
    index: any

    def __repr__(self):
        return f"({self.array}[{self.index}])"
    
@dataclass
class IndexAssignNode:
    array: any
    index: any
    value: any

    def __repr__(self):
        return f"({self.array}[{self.index}] = {self.value})"
    
@dataclass
class VariableNode:
    identifier: str

    def __repr__(self):
        return (f"({self.identifier})")
    
@dataclass
class VarDeclNode:
    identifier: any
    value: any = None

    def __repr__(self):
        return (f"(var {self.identifier} = {self.value})")
    
@dataclass
class AssignNode:
    identifier: any
    value: any

    def __repr__(self):
        return (f"({self.identifier} = {self.value})")
    
@dataclass
class IfNode:
    condition: any
    body: list
    else_body: list = None

    def __repr__(self):
        return (f"(IF {self.condition} THEN \n\t{self.body} \n\tELSE {self.else_body})")

@dataclass
class WhileNode:
    condition: any
    body: list

    def __repr__(self):
        return (f"(WHILE {self.condition} DO {self.body})")

@dataclass
class ForNode:
    init: any
    condition: any
    increment: any
    body: list

    def __repr__(self):
        return (f"(FOR {self.init}; {self.condition}; {self.increment} DO {self.body})")
    
@dataclass
class ParameterNode:
    params: any

    def __repr__(self):
        return f"({self.params})"
    
@dataclass
class ArgumentNode:
    args: any

    def __repr__(self):
        return f"({self.args})"
        
@dataclass
class FunctionDefNode:
    name: str
    params: ParameterNode
    body: any

    def __repr__(self):
        return (f"(FUNCTION {self.name}({self.params}) DO {self.body})")
    
@dataclass
class FunctionCallNode:
    name: str
    args: list

    def __repr__(self):
        return (f"(CALL {self.name.value}({self.args}))")
    
@dataclass
class MethodCallNode:
    object_expr: any
    method_name: any
    args: any

    def __repr__(self):
        return (f"({self.object_expr}.{self.method_name}({self.args}))")
    
@dataclass
class ReturnNode:
    value: any

    def __repr__(self):
        return (f"RETURN {self.value}")


from dataclasses import dataclass
from enum import Enum

WHITESPACE = ' \n\t'
DIGITS = '0123456789.'
KEYWORDS = ["var", "if", "else", "while", "for", "end", "return", "fn", "Nothing", "array"]
LETTERS_LOWER = 'abcdefghijklmnopqrstuvwxyz'
IDENTIFIERS_CHARS = '01234567890abcdefghijklmnopqrstuvwxyz_ABCDEFGHIJKLMNOPQRSTUVWXYZ'
RESERVED_SYMBOLS = '"''=\\;,.<>=^():+-*/!^%[]}{'

class TokenType(Enum):
    INT = 0
    FLOAT = 1
    STRING = 2
    BOOL = 4

    PLUS = 5
    MINUS = 6
    MUL = 7
    DIV = 8
    LPAREN = 9
    RPAREN = 10
    SEMICOLON = 11

    DOT = 12
    VAR = 13
    IDENTIFIER = 14
    EQUAL = 15

    IF = 16
    ELSE = 17
    WHILE = 18
    FOR = 19
    TRUE = 20
    FALSE = 21

    LBRACE = 22
    RBRACE = 23
    COMMA = 24
    COLON = 25

    EQEQ = 26 # ==
    NOT = 27 # !
    GT = 28 # >
    LT = 29 # <
    GTEQ = 30 # >=
    LTEQ = 31 # <=
    NOTEQ = 32
    
    AND = 33
    OR = 34 #
    END = 35
    RETURN = 36

    PLUSPLUS = 37
    MINUSMINUS = 38

    FUNC = 39
    NOTHING = 40

    LBRACKET = 41
    RBRACKET = 42
    ARRAY = 43

@dataclass
class Token:
    type: TokenType
    value: any = None

    def __repr__(self):
        return f"{self.type.name}: {self.value}"
    
class Lexer:
    def __init__(self, text):
        self.text = iter(text)
        self.advance()

    def raise_error(self, message):
        raise Exception(f"{message}")

    def skip_whitespace(self):
        while self.current_char and self.current_char in WHITESPACE:
            self.advance()

    def skip_comments(self):
        self.advance()
        while self.current_char not in ('/', '\n', None):
            self.advance()

        if self.current_char in ('\n', None):
            self.advance()
            return
        elif self.current_char == '/':
            self.advance()
            if self.current_char == '/':
                self.advance()
                return
            else:
                self.skip_comments()
        else:
            self.skip_comments()
            # self.raise_error("Comments left open")

    def advance(self):
        try:
            self.current_char = next(self.text)
            
        except Exception:
            self.current_char = None

    def generate_tokens(self):
        while self.current_char != None:
            if self.current_char in WHITESPACE:
                self.skip_whitespace()
            
            #BINOPS
            elif self.current_char.isdigit() or self.current_char ==  '.': # INTS/FLOATS AND DECIMALS POINTS
                yield self.generate_number()
            # elif self.current_char == '"':
            #     yield Token(TokenType.STRING, self.generate_number(self.current_char))
            elif self.current_char == '+':
                yield self.determine_plus()
            elif self.current_char == '-':
                yield self.determine_minus()
            elif self.current_char == '*':
                yield Token(TokenType.MUL, self.current_char)
                self.advance()
            elif self.current_char == '/': # HANDLES COMMENTS
                self.advance()
                if self.current_char == "/":
                    self.skip_comments()
                elif self.current_char == None:
                    self.raise_error(f"{self.current_char} is not recognised. Did you mean '//'")
                else:
                    yield Token(TokenType.DIV, "/")

            elif self.current_char == '(':
                yield Token(TokenType.LPAREN, self.current_char)
                self.advance()
            elif self.current_char == ')':
                yield Token(TokenType.RPAREN, self.current_char)
                self.advance()
            
            # SYMBOLS + BRACES
            elif self.current_char == ',':
                yield Token(TokenType.COMMA, self.current_char)
                self.advance()
            elif self.current_char == '{':
                yield Token(TokenType.LBRACE, self.current_char)
                self.advance()
            elif self.current_char == '}':
                yield Token(TokenType.RBRACE, self.current_char)
                self.advance()
            elif self.current_char == '[':
                yield Token(TokenType.LBRACKET, self.current_char)
                self.advance()
            elif self.current_char == ']':
                yield Token(TokenType.RBRACKET, self.current_char)
                self.advance()
            elif self.current_char == ';':
                yield Token(TokenType.SEMICOLON, self.current_char)
                self.advance()
            elif self.current_char == ':':
                yield Token(TokenType.COLON, self.current_char)
                self.advance()

            # EQUALS AND CONDITIONAL OP
            elif self.current_char == '=':
                yield self.determine_equals() # NO SELF.ADVANCE()
            elif self.current_char == '<':
                yield self.determine_less_than()
            elif self.current_char == '>':
                yield self.determine_greater_than()
            elif self.current_char == '&':
                yield self.determine_and()
                self.advance()
            elif self.current_char == '|':
                yield self.determine_or()
                self.advance()
            elif self.current_char == '!':
                self.advance()
                if self.current_char == '=':
                    yield Token(TokenType.NOTEQ, '!=')
                    self.advance()
                else:
                    yield Token(TokenType.NOT, '!')

            # STRINGS
            elif self.current_char == '"' or self.current_char == "'":
                yield self.generate_string(self.current_char)

            # IDENTIFIERS & BOOLEANS & KEYWORDS
            elif isinstance(self.current_char, str) and self.current_char in IDENTIFIERS_CHARS:
                yield self.decipher_term()

            # KEYWORDS
            elif isinstance(self.current_char, str) and self.current_char in LETTERS_LOWER:
                yield self.generate_keyword() # CHECK FOR KEYWORDS(E.G. VAR)
            else:
                self.raise_error(r"Invalid syntax")

    def determine_minus(self):
        current_symbol = self.current_char
        self.advance()
        
        if self.current_char == '-':
            current_symbol += self.current_char
            self.advance()
            return Token(TokenType.MINUSMINUS, current_symbol)
        else:
            return Token(TokenType.MINUS, current_symbol)

    def determine_plus(self):
        current_symbol = self.current_char
        self.advance()
        
        if self.current_char == '+':
            current_symbol += self.current_char
            self.advance()
            return Token(TokenType.PLUSPLUS, current_symbol)
        else:
            return Token(TokenType.PLUS, current_symbol)

    def determine_and(self):
        current_symbol = self.current_char
        self.advance()

        if self.current_char == '&':
            current_symbol += self.current_char
            return Token(TokenType.AND, current_symbol)
        else:
            self.raise_error("Invalid Token")

    def determine_or(self):
        current_symbol = self.current_char
        self.advance()

        if self.current_char == '|':
            current_symbol += self.current_char
            return Token(TokenType.OR, current_symbol)
        else:
            self.raise_error("Invalid Token")

    def determine_less_than(self):
        current_less_than = self.current_char
        self.advance()

        if self.current_char == '=':
            current_less_than += self.current_char
            self.advance()
            return Token(TokenType.LTEQ, current_less_than)
        else:
            return Token(TokenType.LT, current_less_than)
        
    def determine_greater_than(self):
        current_more_than = self.current_char
        self.advance()

        if self.current_char == '=':
            current_more_than += self.current_char
            self.advance()
            return Token(TokenType.GTEQ, current_more_than)
        else:
            return Token(TokenType.GT, current_more_than)

    def determine_equals(self):
        current_eq = self.current_char
        self.advance()

        if self.current_char == '=':
            current_eq += self.current_char
            self.advance()
            return Token(TokenType.EQEQ, current_eq)
        else:
            return Token(TokenType.EQUAL, current_eq)
        
    ########################################################

    def generate_number(self):
        current_num = self.current_char
        dp_count = 0
        if self.current_char == '.':
            dp_count += 1
            self.advance()
            if str(self.current_char) not in DIGITS:
                return Token(TokenType.DOT, ".")
        else:
            self.advance()

        while str(self.current_char) in DIGITS and self.current_char != None:
            if str(self.current_char) in DIGITS:
                if self.current_char == '.':
                    dp_count += 1

                current_num = f"{current_num}{self.current_char}"
                self.advance()

        if dp_count <= 1:
            if dp_count > 0:
                if current_num.endswith('.') :
                    current_num = f"{current_num}{0}"
                    return Token(TokenType.FLOAT, float(current_num))
                elif current_num.startswith('.'):
                    current_num = f"{0}{current_num}"
                    return Token(TokenType.FLOAT, float(current_num))
                else:
                    return Token(TokenType.FLOAT, float(current_num))
            return Token(TokenType.INT, int(current_num))
        elif dp_count > 1:
            self.raise_error(r"Too many decimal points")

    def generate_bool(self, value):
        if value == 'True':
            return Token(TokenType.BOOL, True)
        else:
            return Token(TokenType.BOOL, False)
        
    def generate_string(self, quote_type): # INCLUDES QUOTATIONS
        if quote_type == "'" or quote_type == '"':
            current_string = self.current_char
            escape = False
            self.advance()
            
            while self.current_char is not None and (self.current_char != quote_type or escape):
                if escape:
                    if self.current_char == 'n':
                        current_string += '\n'
                    elif self.current_char == 't':
                        current_string += '\t'
                    elif self.current_char == 'b':
                        current_string += '\b'
                    elif self.current_char == '"':
                        current_string += '"'
                    elif self.current_char == "'":
                        current_string += "'"
                    elif self.current_char == '\\':
                        current_string += '\\'
                    escape = False
                else:
                    if self.current_char == '\\':
                        escape = True
                    else:
                        current_string += self.current_char

                self.advance()

            if self.current_char == quote_type:
                current_string = f"{current_string}{self.current_char}" # ADD LAST QUOTATION MARK
                self.advance()

                return Token(TokenType.STRING, str(current_string[1:-1]))
            else:
                self.raise_error(f"Expected ({quote_type}), got ({self.current_char})")
        else:
            self.raise_error("Missing quotations in string")
        
    def generate_keyword(self, keyword=TokenType, value=any):
        return Token(keyword, value)

    def generate_identifier(self, identifier):
        return Token(TokenType.IDENTIFIER, identifier)

    def decipher_term(self):
        if isinstance(self.current_char, str) and self.current_char != None and self.current_char not in RESERVED_SYMBOLS and self.current_char not in WHITESPACE:
            current_word = self.current_char
            self.advance()

        while isinstance(self.current_char, str) and self.current_char != None and self.current_char not in RESERVED_SYMBOLS and self.current_char not in WHITESPACE:
            current_word += self.current_char
            self.advance()
        
        # KEYWORDS > BOOLEAN > IDENTIFIER
        if current_word in KEYWORDS:
            if current_word == 'var':
                return self.generate_keyword(TokenType.VAR, current_word)
            elif current_word == 'array':
                return self.generate_keyword(TokenType.ARRAY, current_word)
            elif current_word == 'if':
                return self.generate_keyword(TokenType.IF, current_word)
            elif current_word == 'else':
                return self.generate_keyword(TokenType.ELSE, current_word)
            elif current_word == 'while':
                return self.generate_keyword(TokenType.WHILE, current_word)
            elif current_word == 'for':
                return self.generate_keyword(TokenType.FOR, current_word)
            elif current_word == 'return':
                return self.generate_keyword(TokenType.RETURN, current_word)
            elif current_word == 'end':
                return self.generate_keyword(TokenType.END, current_word)
            elif current_word == 'fn':
                return self.generate_keyword(TokenType.FUNC, current_word)
            elif current_word == 'Nothing':
                return self.generate_keyword(TokenType.NOTHING, current_word)
        elif current_word in ("True", "False"):
            return self.generate_bool(current_word)
        else:
            return self.generate_identifier(current_word)


import time
from lexer import Lexer
from parser import Parser
from runtime.interpreter import Interpreter

current_ver = "0.1.0"
start_time = time.perf_counter()

def run_file(file_path):
    try:
        with open(file_path, 'r') as file:
            content = file.read()
            
        lexer = Lexer(content)
        tokens = list(lexer.generate_tokens())
        # print(tokens)
        parser = Parser(tokens)
        tree = parser.parse()
        # print(tree)

        interpreter = Interpreter()
        interpreter.interpret(tree)
        # print(interpreter.env.scopes)
    except FileNotFoundError:
        raise Exception('file does not exist')

if __name__ == '__main__':
    pass

end_time = time.perf_counter()
# print(f"Executed in {(end_time-start_time)*1000:.4f}ms")


from lexer import TokenType, Token
from tree import *

class Parser:
    def __init__(self, tokens):
        self.tokens = tokens or []
        self.pos = 0
        if tokens:
            self.current_token = self.tokens[self.pos]

    def raise_error_expect(self, expected, got=None):
        if got == None:
            raise Exception(f"Expected '{expected}'")
        else:
            raise Exception(f"Expected '{expected}', got '{got.value}'")
        
    def raise_error(self, message):
        if message:
            raise Exception(str(message))
        else:
            raise Exception

    def advance(self):
        self.pos += 1
        if self.current_token and self.pos < len(self.tokens):
            self.current_token = self.tokens[self.pos]
        else:
            self.current_token = None

    def peek(self):
        if self.pos + 1 < len(self.tokens) and self.pos + 1 != None:
            return self.tokens[self.pos + 1]
        return None
    
    def peek_prev_token(self, count=1):
        if self.pos - 1 >= 0:
            return self.tokens[self.pos - count]
        return None

    def parse(self):
        statements = []
        
        if self.tokens:
            while self.current_token != None:
                stmt = self.parse_statement()
                if stmt:
                    statements.append(stmt)
                
                # HANDLE SEMICOLONS/LINE ENDINGS
                if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                    self.advance()
                elif self.peek_prev_token() and self.peek_prev_token().type == TokenType.END:
                    continue
                else:
                    self.raise_error_expect(';', self.current_token)

        return ProgramNode(statements)

    def parse_statement(self):
        token = self.current_token

        if token.type == TokenType.IDENTIFIER and self.peek():
            return self.parse_identifier()
        if token.type == TokenType.VAR:
            return self.parse_var_decl()
        elif token.type == TokenType.ARRAY:
            return self.parse_array_decl()
        elif token.type == TokenType.IF:
            return self.parse_if_statement()
        elif token.type == TokenType.WHILE:
            return self.parse_while_statement()
        elif token.type == TokenType.FOR:
            return self.parse_for_statement()
        elif token.type == TokenType.FUNC:
            return self.parse_function_def()
        elif token.type == TokenType.RETURN:
            return self.parse_return_statement()
        else:
            # if token and token.type in (TokenType.INT, TokenType.FLOAT, TokenType.STRING, TokenType.MUL, TokenType.DIV, TokenType.PLUS, TokenType.MINUS, TokenType.LPAREN, TokenType.RPAREN):
            #     return self.parse_expr()
            return self.raise_error(f"Unexpected token: {token.value}")
        
    def parse_identifier(self):
        peek = self.peek()
        if peek != None and peek.type in (TokenType.EQUAL, TokenType.LBRACKET):
            return self.parse_assignment()
        elif peek != None and peek.type == TokenType.LPAREN:
            return self.parse_function_call()
        elif peek != None and peek.type == TokenType.DOT:
            return self.parse_method_call()
        else:
            self.raise_error(f"Invalid token after: {self.current_token}")
    
    def parse_assignment(self):
        identifier = self.parse_index_access()
        if self.current_token.type == TokenType.EQUAL:
            self.advance()
            if isinstance(identifier, IndexAccessNode):
                return IndexAssignNode(identifier.array, identifier.index, self.parse_expr())
            return AssignNode(getattr(identifier, 'value', identifier), self.parse_expr())

    def parse_var_decl(self):
        self.advance()
        if self.current_token and self.current_token.type == TokenType.IDENTIFIER:
            identifier = self.current_token.value
            self.advance()

            if self.current_token and self.current_token.type == TokenType.EQUAL:
                self.advance()
                value = self.parse_expr()
                if value:
                    return VarDeclNode(identifier, value)
                self.raise_error(f"'{identifier}' must have a declared value")
            else:
                self.raise_error_expect("=", self.current_token)
        else:
            self.raise_error_expect("identifier", self.current_token)

    def parse_array(self):
        elements = []
        self.advance()
        if self.current_token and self.current_token.type != TokenType.RBRACKET:
            stmt = self.parse_expr()
            if stmt:
                elements.append(stmt)
            while self.current_token and self.current_token.type == TokenType.COMMA:
                self.advance()
                stmt = self.parse_expr()
                if stmt:
                    elements.append(stmt)
            if self.current_token and self.current_token.type == TokenType.RBRACKET:
                self.advance()
                return elements
            else:
                self.raise_error_expect("]", self.current_token)

    def parse_array_decl(self):
        self.advance()
        if self.current_token and self.current_token.type == TokenType.IDENTIFIER:
            identifier = self.current_token.value
            self.advance()
            if self.current_token and self.current_token.type == TokenType.EQUAL:
                self.advance()
                if self.current_token and self.current_token.type == TokenType.LBRACKET:
                    elements = self.parse_array()
                    return VarDeclNode(identifier, ArrayNode(elements))
                else:
                    self.raise_error_expect("[", self.current_token)
            else:
                self.raise_error_expect("=", self.current_token)
        else:
            self.raise_error_expect("identifier", self.current_token)

    def parse_index_access(self):
        node = self.parse_primary()
        while self.current_token and self.current_token.type == TokenType.LBRACKET:
            self.advance()
            index = self.parse_expr()
            if self.current_token and self.current_token.type == TokenType.RBRACKET:
                self.advance()
                node = IndexAccessNode(node, index)
            else:
                self.raise_error_expect("]", self.current_token)
        return node
    
    def parse_primary(self):
        token = self.current_token
        if token.type == TokenType.IDENTIFIER:
            self.advance()
            return VariableNode(token.value)
        elif token.type in (TokenType.INT, TokenType.FLOAT, TokenType.STRING, TokenType.BOOL):
            self.advance()
            return LiteralNode(token)
        elif token.type == TokenType.NOTHING:
            self.advance()
            return NothingLiteralNode(None)
        elif token.type == TokenType.LPAREN:
            self.advance()
            expr = self.parse_expr()
            if self.current_token and self.current_token.type == TokenType.RPAREN:
                self.advance()
                return expr
            else:
                self.raise_error_expect(")", self.current_token)
        else:
            self.raise_error(f"Unexpected token in primary: {token}")

    def parse_return_statement(self):
        self.advance()
        if self.current_token and self.current_token.type in (TokenType.SEMICOLON, TokenType.NOTHING):
            self.advance()
            return ReturnNode(NothingLiteralNode(None))
        value = self.parse_expr()
        return ReturnNode(value)

    # FUNCTIONS AND METHODS
    def parse_method_call(self):
        node = VariableNode(self.current_token.value)
        self.advance()
        while self.current_token and self.current_token.type == TokenType.DOT:
            self.advance()
            method_name = self.current_token.value
            self.advance()
            self.advance()
            args = self.parse_arguments()
            if self.current_token and self.current_token.type == TokenType.RPAREN:
                self.advance()
            else:
                self.raise_error_expect(")", self.current_token)
            node = MethodCallNode(node, method_name, args)
            if args and args == "error":
                self.raise_error(f"Unexpected argument error for method '{method_name}()'")
        return node

    def parse_function_call(self):
        name = self.current_token
        self.advance()
        if self.current_token and self.current_token.type == TokenType.LPAREN:
            self.advance()
            args = self.parse_arguments()
            if args and args == "error":
                self.raise_error(f"Unexpected argument error for method '{name.value}()'")
            if self.current_token and self.current_token.type == TokenType.RPAREN:
                self.advance()
                return FunctionCallNode(name, args)
            else:
                self.raise_error_expect(")", self.current_token)
        else:
            self.raise_error_expect("(", self.current_token)

    def parse_arguments(self):
        if self.current_token and self.peek_prev_token() and self.peek_prev_token().type == TokenType.LPAREN:
            try:
                args = []
                if self.current_token and self.current_token.type in (TokenType.IDENTIFIER, TokenType.STRING, TokenType.INT, TokenType.FLOAT, TokenType.BOOL, TokenType.LBRACKET, TokenType.LPAREN):
                    args.append(self.parse_expr())
                    while self.current_token and self.current_token.type == TokenType.COMMA:
                        if self.current_token and self.current_token.type == TokenType.COMMA:
                            self.advance()
                            if self.current_token and self.current_token.type not in (TokenType.IDENTIFIER, TokenType.INT, TokenType.FLOAT, TokenType.STRING, TokenType.BOOL, TokenType.NOTHING):
                               return "error"
                        if self.current_token and self.current_token.type != TokenType.COMMA:
                            args.append(self.parse_expr())
                return args
            except Exception:
                return "error"
        else:
            self.raise_error_expect("(", self.current_token)

    def parse_function_def(self):
        self.advance()
        name = None
        params = []
        body = []
        if self.current_token and self.current_token.type == TokenType.IDENTIFIER:
            name = self.current_token.value
            self.advance()
            if self.current_token and self.current_token.type == TokenType.LPAREN:
                self.advance()
                if self.current_token and self.current_token.type != TokenType.RPAREN:
                    params = self.parse_parameters()
                
                if params and params == "error":
                    self.raise_error(f"Unexpected parameter error for function '{name.value}()'")
                if self.current_token and self.current_token.type == TokenType.RPAREN:
                    self.advance()
                else:
                    self.raise_error_expect(")", self.current_token)

                if self.current_token and self.current_token.type == TokenType.COLON:
                    self.advance()
                    while self.current_token and self.current_token.type != TokenType.END:
                        stmt = self.parse_statement()
                        if stmt:
                            body.append(stmt)
                        if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                            self.advance()
                        elif self.peek_prev_token() and self.peek_prev_token().type == TokenType.END:
                            continue
                        elif self.current_token and self.current_token.type == TokenType.END:
                            break
                        else:
                            self.raise_error_expect(";", self.current_token)
                    if self.current_token and self.current_token.type == TokenType.END:
                        self.advance()
                    else:
                        self.raise_error_expect("end", self.current_token)
                else:
                    self.raise_error_expect(":", self.current_token)
            else:
                self.raise_error_expect("(", self.current_token)
        else:
            self.raise_error("Expected function identifier")

        return FunctionDefNode(name, params, body)
    
    def parse_parameters(self):
        if self.current_token and self.peek_prev_token() and self.peek_prev_token().type == TokenType.LPAREN:
            try:
                params = []
                if self.current_token and self.current_token.type == TokenType.IDENTIFIER:
                    params.append(ParameterNode(self.parse_expr()))
                    while self.current_token and self.current_token.type == TokenType.COMMA:
                        if self.current_token and self.current_token.type == TokenType.COMMA:
                            self.advance()
                        if self.current_token and self.current_token.type != TokenType.COMMA:
                            params.append(ParameterNode(self.parse_expr()))
                return params
            except Exception:
                return "error"
        else:
            self.raise_error_expect("(", self.current_token)
            
    ############### CONDITIONALS AND LOOPS #################
    def parse_if_statement(self):
        self.advance()
        if self.current_token.type == TokenType.LPAREN:
            self.advance()
            condition = self.parse_logical_or() ### <- FIX HERE

            if self.current_token.type == TokenType.RPAREN:
                self.advance()
            else:
                self.raise_error_expect(")", self.current_token)

            # BODY
            if self.current_token.type == TokenType.COLON:
                self.advance()
                body = []
                while self.current_token.type not in (TokenType.END, TokenType.ELSE):
                    start_pos = self.pos
                    stmt = self.parse_statement()
                    if stmt:
                        body.append(stmt)
                    if start_pos == self.pos:
                        self.raise_error("Parser did not advance in if statement")
                    # SEMICOLON
                    if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                        self.advance()
                    elif self.current_token and self.current_token.type == TokenType.END:
                        break
                    elif self.peek_prev_token() and self.peek_prev_token().type == TokenType.END:
                        continue
                    else:
                        self.raise_error_expect(";")
                        
                if self.current_token and self.current_token.type == TokenType.ELSE:
                    else_body = []
                    self.advance()
                    if self.current_token and self.current_token.type == TokenType.IF:
                        stmt = self.parse_if_statement()
                        if stmt:
                            else_body.append(stmt)
                    else:
                        if self.current_token and self.current_token.type == TokenType.COLON:
                            self.advance()
                            while self.current_token and self.current_token.type != TokenType.END:
                                stmt = self.parse_statement()
                                if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                                    self.advance()
                                else:
                                    self.raise_error_expect(";")
                                if stmt:
                                    else_body.append(stmt)
                            if self.current_token and self.current_token.type == TokenType.END:
                                self.advance()
                                if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                                    self.advance()
                        else:
                            self.raise_error_expect(":")

                    if else_body == []:
                        self.raise_error("Expected content inside of else")
                    
                    return IfNode(condition, body, else_body)
                elif self.current_token and self.current_token.type == TokenType.END:
                    self.advance()
                    if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                        self.advance()
                    return IfNode(condition=condition, body=body)
            else:
                self.raise_error_expect(expected=":", got=self.current_token)

    def parse_while_statement(self):
        self.advance()
        if self.current_token and self.current_token.type == TokenType.LPAREN:
            self.advance()
            condition = self.parse_logical_or()
            if self.current_token and self.current_token.type == TokenType.RPAREN:
                self.advance()
            else:
                self.raise_error_expect(")")

            # BODY
            if self.current_token.type == TokenType.COLON:
                self.advance()
                body = []
                while self.current_token and self.current_token.type != TokenType.END:
                    stmt = self.parse_statement()
                    if stmt:
                        body.append(stmt)
                    if self.current_token.type == TokenType.SEMICOLON:
                        self.advance()
                    else:
                        self.raise_error_expect(";")
                if self.current_token and self.current_token.type == TokenType.END:
                    self.advance()
                    return WhileNode(condition, body)
                else:
                    self.raise_error_expect("end")
            else:
                self.raise_error_expect(":")
        else:
            self.raise_error_expect("(")

    def parse_for_statement(self):
        self.advance()
        if self.current_token and self.current_token.type == TokenType.LPAREN:
            self.advance()
            if self.current_token and self.current_token.type == TokenType.VAR:
                init = self.parse_var_decl()
                if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                    self.advance()
                    condition = self.parse_logical_or()
                    if self.current_token and self.current_token.type == TokenType.SEMICOLON:
                        self.advance()
                        # POST DECREMENT/INCREMENT
                        if self.current_token and self.current_token.type == TokenType.IDENTIFIER: 
                            postfixop_identifier = self.current_token
                            self.advance()
                            if self.current_token.type in (TokenType.MINUSMINUS, TokenType.PLUSPLUS):
                                postfixop_op = self.current_token
                                self.advance()

                                # RPAREN
                                if self.current_token and self.current_token.type == TokenType.RPAREN:
                                    self.advance()
                                    unopchange = PostfixOpNode(postfixop_identifier.value, postfixop_op)
                                else:
                                    self.raise_error_expect(")", self.current_token)
                            elif self.current_token and self.current_token.type == TokenType.EQUAL:
                                self.advance()
                                unopchange = AssignNode(postfixop_identifier.value, self.parse_expr())
                                if self.current_token and self.current_token.type == TokenType.RPAREN:
                                    self.advance()
                                else:
                                    self.raise_error_expect(")", self.current_token)
                            else:
                                self.raise_error("Expected assignment in for statement")
                        # PRE-DECREMENT/INCREMENT
                        elif self.current_token and self.current_token.type in (TokenType.MINUSMINUS, TokenType.PLUSPLUS): #Prefix Op
                            op = self.current_token
                            self.advance()
                            if self.current_token and self.current_token.type == TokenType.IDENTIFIER:
                                identifier = self.current_token
                                self.advance()
                                # RPAREN
                                if self.current_token and self.current_token.type == TokenType.RPAREN:
                                    self.advance()
                                    unopchange = AssignNode(identifier.value, BinaryOpNode(VariableNode(f"{identifier.value}"), Token(TokenType.PLUS, "+"), LiteralNode(Token(TokenType.INT, 1))))
                                else:
                                    self.raise_error_expect(")")
                            else:
                                self.raise_error(f"Expected identifier after '{op.value}'")
                        else:
                            self.raise_error("Expected increment or decrement")
                    else:
                        self.raise_error_expect(";")
                else:
                    self.raise_error_expect(";")
            else:
                self.raise_error_expect("var")
        else:
            self.raise_error_expect(")")
        
        # BODY
        if self.current_token.type == TokenType.COLON:
            self.advance()
            body = []
            while self.current_token != None and self.current_token.type not in (TokenType.END, None):
                stmt = self.parse_statement()
                if stmt:
                    body.append(stmt)
                if self.current_token.type == TokenType.SEMICOLON:
                    self.advance()
                elif self.peek_prev_token() and self.peek_prev_token().type == TokenType.END and self.current_token.type != TokenType.SEMICOLON:
                    continue                   
                else:
                    self.raise_error_expect(";")
            if body == []:
                self.raise_error("Body cannot be empty")
            if self.current_token and self.current_token.type == TokenType.END:
                self.advance()
            else:
                self.raise_error_expect("end")
        else:
            self.raise_error_expect(":")

        return ForNode(init, condition, unopchange, body)
    
    def parse_logical_or(self):
        left = self.parse_logical_and()
        while self.current_token != None and self.current_token.type == TokenType.OR:
            op = self.current_token
            self.advance()
            right = self.parse_logical_and()
            left = BinaryOpNode(left, op, right)
        return left
    
    def parse_logical_and(self):
        left = self.parse_condition()
        while self.current_token != None and self.current_token.type == TokenType.AND:
        
            op = self.current_token
            self.advance()
            right = self.parse_condition()
            left = BinaryOpNode(left, op, right)
        return left
    
    def parse_condition(self):
        left = self.parse_expr()
        while self.current_token != None and self.current_token.type in (TokenType.LT, TokenType.GT, TokenType.EQEQ, TokenType.LTEQ, TokenType.GTEQ, TokenType.NOTEQ):
            op = self.current_token
            self.advance()
            right = self.parse_expr()
            left = BinaryOpNode(left, op, right)
        if self.current_token.type == TokenType.EQUAL:
            self.raise_error("Unexpected operator")
        return left

    def parse_expr(self):
        left = self.parse_term()

        while self.current_token != None and self.current_token.type in (TokenType.PLUS, TokenType.MINUS): 
            op = self.current_token
            self.advance()
            right = self.parse_term()
            left = BinaryOpNode(left, op, right)
        return left
    
    def parse_term(self):
        left = self.parse_factor()
        while self.current_token != None and self.current_token.type in (TokenType.MUL, TokenType.DIV):
            op = self.current_token
            self.advance()
            right = self.parse_factor()
            left = BinaryOpNode(left, op, right)
        return left
    
    def parse_factor(self):
        token = self.current_token

        if token.type in (TokenType.PLUS, TokenType.MINUS):
            self.advance()
            op = token
            operand = self.parse_factor()
            return UnaryOpNode(op, operand)
        elif token.type == TokenType.LPAREN:
            self.advance()
            expr = self.parse_expr()
            if self.current_token and self.current_token.type == TokenType.RPAREN:
                self.advance()
                return expr
            else:
                self.raise_error_expect(")", self.current_token)
        elif token.type == TokenType.LBRACKET:
            array = self.parse_array()
            return ArrayNode(array)
        elif token.type == TokenType.IDENTIFIER:
            if self.peek() and self.peek().type == TokenType.LPAREN:
                return self.parse_function_call()
            elif self.peek() and self.peek().type == TokenType.LBRACKET:
                return self.parse_index_access()
            else:
                return self.parse_index_access()  # handles simple var or chained array access
        elif token.type in (TokenType.INT, TokenType.FLOAT, TokenType.STRING, TokenType.BOOL):
            self.advance()
            return LiteralNode(token)
        elif token.type == TokenType.NOTHING:
            self.advance()
            return NothingLiteralNode(None)
        else:
            self.raise_error(f"Unexpected token in factor: {token}")


from dataclasses import dataclass

@dataclass
class BuiltInFunctionValue:
    name: any
    func: any

    def call(self, interpreter, args):
        return self.func(interpreter, args)

    def __repr__(self):
        return (f"<builtin {self.name}>")

@dataclass
class IntValue:
    value: int

    def __repr__(self):
        return (f"{self.value}")

@dataclass
class FloatValue:
    value: float

    def __repr__(self):
        return (f"{self.value}")

@dataclass
class StringValue:
    value: float

    def __repr__(self):
        return (f"{self.value}")
    
@dataclass
class BoolValue:
    value: float

    def __repr__(self):
        return (f"{self.value}")
    
@dataclass
class VariableValue:
    value: any

class NothingValue:
    def __init__(self, value=None):
        self.value = value
        if value == None:
            self.value = 'Nothing'

    def __repr__(self):
        return (f"{self.value}")
    
@dataclass
class FunctionValue:
    name: str
    params: any
    body: any
    env: any

class ReturnSignal(Exception):
    def __init__(self, value):
        self.value = value


from lexer import TokenType
from runtime.values_ import *
from tree import *
from runtime.environment import Environment
import sys
# from utils.errors import *

class Interpreter:
    def __init__(self):
        self.env = Environment()
        self.load_stdlib()

    def raise_error(self, message):
        raise Exception(message)

    def interpret(self, tree):
        if isinstance(tree, ProgramNode):
            self.visit(tree)

    def visit(self, node):
        method_name = f'visit_{node.__class__.__name__}'
        method = getattr(self, method_name, self.no_visit_method)
        return method(node)

    def no_visit_method(self, node):
        raise Exception(f"No visit_{type(node).__name__} method defined")
    
    def visit_ProgramNode(self, node):
        result = None

        for statement in node.body:
            result = self.visit(statement)

        return result
    
    def visit_IndexAccessNode(self, node):
        arr = self.visit(node.array)
        index = self.visit(node.index).value
        if not isinstance(arr, list):
            self.raise_error("Cannot index non-list")
        if not isinstance(index, int):
            self.raise_error("Cannot index with non-integer")
        if abs(index) > len(arr):
            self.raise_error("Index is out of range")

        return arr[index]
    
    def visit_IndexAssignNode(self, node):
        arr = self.visit(node.array)
        index = self.visit(node.index).value
        value = self.visit(node.value)

        if not isinstance(arr, list):
            self.raise_error("Cannot index non-list")
        if not isinstance(index, int):
            self.raise_error("Cannot index with non-integer")
        if abs(index) > len(arr):
            self.raise_error("Index is out of range")

        arr[index] = value
        return value
    
    def visit_MethodCallNode(self, node):
        def handle_array_method(arr, method, args):
            match(method):
                case 'push':
                    if len(args) < 1:
                        self.raise_error("push() method takes at least 1 argument")
                    
                    for i in args:
                        arr.append(i)
                
                case 'pop':
                    if len(args) > 0:
                        self.raise_error("pop() method takes no arguments")
                    arr.pop()

                case 'insert':
                    if len(args) != 2:
                        self.raise_error("insert() method takes exactly 2 arguments: (index, value)")

                    idx, value = args
                    
                    if not isinstance(idx.value, int):
                        self.raise_error("insert(): index argument must be an integer")
                    if idx.value < 0 or idx.value >= len(arr):
                        self.raise_error("insert(): index argument is out of bounds")

                    arr.insert(idx.value, value)

                case 'delete':
                    if len(args) != 1:
                        self.raise_error("delete() method takes exactly 1 argument")
                    
                    idx = args[0]
                    
                    if not isinstance(idx.value, int):
                        self.raise_error("delete(): index argument must be an integer")
                    if idx.value < 0 or idx.value >= len(arr):
                        self.raise_error("delete(): index argument is out of bounds")

                    arr.pop(idx.value)

                case _:
                    self.raise_error(f"Unknown array method '{method}'")

        obj = self.visit(node.object_expr)

        method = node.method_name
        args = [self.visit(arg) for arg in node.args]

        if isinstance(obj, list):
            return handle_array_method(obj, method, args)
        
        raise Exception(f"'{method}' cannot be called on this type")

    def visit_ArrayNode(self, node):
        return [self.visit(element) for element in node.elements]
    
    def visit_ReturnNode(self, node):
        raise ReturnSignal(self.visit(node.value))
    
    def visit_ParameterNode(self, node):
        return node.params.identifier

    def visit_FunctionDefNode(self, node):
        function = FunctionValue(node.name, node.params, node.body, self.env)
        self.env.declare(node.name, function)

    def visit_FunctionCallNode(self, node):
        func = self.env.get(node.name.value)

        # Evaluate arguments in the CURRENT environment (before switching)
        arg_values = [self.visit(arg) for arg in node.args]

        if isinstance(func, BuiltInFunctionValue):
            return func.call(self, arg_values)
        
        call_env = Environment(func.env)
        call_env.enter_scope()

        old_env = self.env
        self.env = call_env
        
        # Declare parameters with pre-evaluated argument values
        for param, arg_value in zip(func.params, arg_values):
            param_name = self.visit(param)
            call_env.declare(param_name, arg_value)
        
        try:
            for stmt in func.body:
                self.visit(stmt)
        except ReturnSignal as r:
            result = r.value
            call_env.exit_scope()
            self.env = old_env
            return result

        call_env.exit_scope()
        self.env = old_env
    
    def visit_WhileNode(self, node):
        self.env.enter_scope()
        while self.visit(node.condition):
            for stmt in node.body:
                self.visit(stmt)
        self.env.exit_scope()
    
    def visit_ForNode(self, node):
        self.env.enter_scope()
        self.visit(node.init)
        while self.visit(node.condition):
            for stmt in node.body:
                self.visit(stmt)
            
            self.visit(node.increment)
        self.env.exit_scope()

    def visit_IfNode(self, node):
        condition = self.visit(node.condition)
        if condition:
            self.env.enter_scope()
            for stmt in node.body:
                self.visit(stmt)
            self.env.exit_scope()
        elif node.else_body:
            self.env.enter_scope()
            for stmt in node.else_body:
                self.visit(stmt)
            self.env.exit_scope()
    
    def visit_AssignNode(self, node):
        value = self.visit(node.value)
        identifier = getattr(node.identifier, 'identifier', node.identifier)
        self.env.assign(identifier, value)

    def visit_VarDeclNode(self, node):
        value = self.visit(node.value)
        self.env.declare(node.identifier, value)

    def visit_UnaryOpNode(self, node):
        operand = self.visit(node.operand)
        op = node.op
        if op.value == '+':
            if isinstance(operand, IntValue):
                return IntValue(operand.value)
            elif isinstance(operand, FloatValue):
                return FloatValue(operand.value)
        elif op.value == '-':
            if isinstance(operand, IntValue):
                return IntValue(operand.value * -1)
            elif isinstance(operand, FloatValue):
                return FloatValue(operand.value * -1)

    def visit_BinaryOpNode(self, node):
        def eval_values(left, op, right):
            left = getattr(left, 'value', left)
            right = getattr(right, 'value', right)
            # Nothing value
            if op in ('+', '-', '*', '/') and None in (left, right):
                raise TypeError(f"Cannot perform operations with 'Nothing' type")
            if op == '+':
                if isinstance(left, str) or isinstance(right, str):
                    return (str(left)+str(right))
                return (left+right)
            elif op == '-': return (left-right)
            elif op == '*': return (left*right)
            elif op == '/':
                if right != 0: return (left/right) 
                else: raise ZeroDivisionError("Cannot divide by zero")
            elif op == '==': return bool(left==right)
            elif op == '>': return bool(left>right)
            elif op == '<': return bool(left<right)
            elif op == '<=': return bool(left<=right)
            elif op == '>=': return bool(left>=right)
            elif op == '!=': return bool(left!=right)

        op = node.op
        left = self.visit(node.left)
        if op.type == TokenType.AND:
            if left == False:
                return False
            else:
                return self.visit(node.right)
        elif op.type == TokenType.OR:
            if left:
                return True
            else:
                return self.visit(node.right)
        right = self.visit(node.right)
        if op.type == TokenType.PLUS:
            return eval_values(left, '+', right)
        elif op.type == TokenType.MINUS:
            return eval_values(left, '-', right)
        elif op.type == TokenType.MUL:
            return eval_values(left, '*', right)
        elif op.type == TokenType.DIV:
            return eval_values(left, '/', right)
        # BOOLEAN OPS
        elif op.type == TokenType.EQEQ:
            return eval_values(left, '==', right)
        elif op.type == TokenType.GT:
            return eval_values(left, '>', right)
        elif op.type == TokenType.LT:
            return eval_values(left, '<', right)
        elif op.type == TokenType.LTEQ:
            return eval_values(left, '<=', right)
        elif op.type == TokenType.GTEQ:
            return eval_values(left, '>=', right)
        elif op.type == TokenType.NOTEQ:
            return eval_values(left, '!=', right)
        
    def visit_VariableNode(self, node):
        identifier = node.identifier
        value = self.env.get(identifier)
        return value
    
    def visit_PostfixOpNode(self, node):
        identifier = node.identifier
        value = self.env.get(identifier)
        if node.op.type == TokenType.PLUSPLUS:
            old_value = value
            value.value += 1
            self.env.assign(identifier, value)
        if node.op.type == TokenType.MINUSMINUS:
            old_value = value
            value.value -= 1
            self.env.assign(identifier, value)
        return old_value
        
    def visit_NothingLiteralNode(self, node):
        token = node.literal
        if token == None:
            return NothingValue(token)
    
    def visit_LiteralNode(self, node):
        token = node.literal
        if token.type == TokenType.INT:
            return IntValue(token.value)
        elif token.type == TokenType.FLOAT:
            return FloatValue(token.value)
        elif token.type == TokenType.STRING:
            return StringValue(token.value)
        elif token.type == TokenType.BOOL:
            return BoolValue(token.value)
        
    def load_stdlib(self):
        def builtin_print(interpreter, args):
            values = [getattr(arg, 'value', arg) for arg in args]
            for val in values:
                if isinstance(val, (FunctionValue, BuiltInFunctionValue)):
                    self.raise_error(f"function {val.name}() has invalid syntax when called")
                print(val)
            return NothingValue(None)
        
        def builtin_int(interpreter, args):
            if not args:
                raise Exception("int() needs 1 argument")
            elif len(args) > 1:
                raise Exception("int() only takes in 1 argument")
            return IntValue(int(getattr(args[0], 'value', args[0])))
        
        def builtin_float(interpreter, args):
            if not args:
                raise Exception("float() needs 1 argument")
            elif len(args) > 1:
                raise Exception("float() only takes in 1 argument")
            return FloatValue(float(getattr(args[0], 'value', args[0])))
        
        def builtin_string(interpreter, args):
            if not args:
                raise Exception("string() needs 1 argument")
            elif len(args) > 1:
                raise Exception("string() only takes in 1 argument")
            return StringValue(str(getattr(args[0], 'value', args[0])))
        
        def builtin_type(interpreter, args):
            if not args:
                raise Exception("string() needs 1 argument")
            elif len(args) > 1:
                raise Exception("string() only takes in 1 argument")
            return type(args)
        
        def builtin_size(interpreter, args):
            if not args:
                raise Exception("size() needs 1 argument")
            elif len(args) > 1:
                raise Exception("size() only takes in 1 argument")
            return len(getattr(args[0], 'value', args[0]))
        
        def builtin_ask(interpreter, args):
            if not args:
                raise Exception("ask() needs 1 argument")
            elif len(args) > 1:
                raise Exception("ask() only takes in 1 argument")
            return input(args[0])
        
        def builtin_terminate(interpreter, args):
            if args:
                raise Exception("terminate() takes exactly no arguments")
            elif len(args) > 1:
                raise Exception("terminate() takes exactly no arguments")
            quit()

        # def builtin_sqrt(interpreter, args):
        #     values = 
        
        self.env.declare("print", BuiltInFunctionValue("print", builtin_print))
        self.env.declare("int", BuiltInFunctionValue("int", builtin_int))
        self.env.declare("float", BuiltInFunctionValue("float", builtin_float))
        self.env.declare("string", BuiltInFunctionValue("string", builtin_string))
        self.env.declare("size", BuiltInFunctionValue("size", builtin_size))
        self.env.declare("ask", BuiltInFunctionValue("ask", builtin_ask))
        self.env.declare("terminate", BuiltInFunctionValue("terminate", builtin_terminate))


import sys
from main import run_file

def main():
    if len(sys.argv) == 2:
        run_file(sys.argv[1])
        sys.exit(1)
    else:
        print("Usage: orion <script.or>")
        sys.exit(1)

main()



class Environment:
    def __init__(self, parent=None):
        self.parent = parent
        self.scopes = [{}]

    def enter_scope(self):
        self.scopes.append({})
    
    def exit_scope(self):
        if len(self.scopes) > 1:
            self.scopes.pop()

    def declare(self, identifier, value):
        current_scope = self.scopes[-1]
        if str(identifier) in current_scope:
            raise Exception(f"'{identifier}' already exists")
        current_scope[str(identifier)] = value

    def assign(self, identifier, value):
        for scope in reversed(self.scopes):
            if identifier in scope:
                scope[identifier] = value
                return
            
        if self.parent:
            self.parent.assign(identifier, value)
            return

        raise Exception(f"'{identifier}' is not declared")
    
    def get(self, identifier):
        for scope in reversed(self.scopes):
            if identifier in scope:
                value = scope.get(identifier)
                return value
            
        if self.parent:
            return self.parent.get(identifier)
        raise Exception(f"'{identifier}' does not exist")
    
    def debug(self):
        print("SCOPE: ", self.scopes)
        if self.parent:
            print("Parent →")
            self.parent.debug()